import { _ as _define_property } from "@swc/helpers/_/_define_property";
import { RangeSetBuilder } from '@codemirror/state';
import { Decoration, EditorView, ViewPlugin, WidgetType } from '@codemirror/view';
import { slug as createSlug } from 'github-slugger';
let SlugWidget = class SlugWidget extends WidgetType {
    toDOM(_view) {
        const el = window.document.createElement('span');
        el.innerHTML = `#${this.renderSlug}`;
        el.classList.add('codemirror-markdown-slug');
        if (!this.slug) {
            el.classList.add('codemirror-markdown-slug--default');
        }
        return el;
    }
    ignoreEvent() {
        return false;
    }
    constructor(title, slug){
        super();
        _define_property(this, "title", void 0);
        _define_property(this, "slug", void 0);
        _define_property(this, "renderSlug", void 0);
        this.title = title;
        this.slug = slug;
        this.renderSlug = slug || createSlug(title.trim());
    }
};
export const TITLE_MATCH_REGEXP = /^#+\s.*/;
const parseTitle = (text)=>{
    const res = {
        level: 0,
        title: '',
        slug: ''
    };
    const match = text.match(RegExp("^(?<level>#{1,6}) (?<content>.*)$"));
    if (!match) {
        return res;
    }
    res.level = match.groups.level.length;
    const titleContent = match.groups.content;
    const slugMatch = titleContent.match(RegExp("^(?<content>.*) \\{#(?<slug>.*)}$"));
    if (slugMatch) {
        res.title = slugMatch.groups.content;
        res.slug = slugMatch.groups.slug;
    } else {
        res.title = titleContent;
    }
    return res;
};
const isHeading = (text)=>{
    return !!text && TITLE_MATCH_REGEXP.test(text);
};
export class SlugPlugin {
    getDecorationForText(line) {
        if (!line.text || !isHeading(line.text)) {
            return;
        }
        const { level, title, slug } = parseTitle(line.text);
        if (!title) {
            return;
        }
        return {
            level,
            title,
            slug,
            from: line.from + title.length + level + 1,
            to: line.to,
            value: Decoration.replace({
                widget: new SlugWidget(title, slug)
            })
        };
    }
    renderWidgets() {
        const { state } = this.view;
        const { doc } = state;
        const currentLine = this.view.hasFocus ? doc.lineAt(state.selection.main.head).number : 0;
        const lineCount = doc.lines;
        const decorations = new RangeSetBuilder();
        for(let i = 1; i <= lineCount; i++){
            if (i === currentLine) {
                continue;
            }
            const line = doc.line(i);
            if (!isHeading(line.text)) {
                continue;
            }
            const item = this.getDecorationForText(line);
            if (item) {
                decorations.add(item.from, item.to, item.value);
            }
        }
        this.decorations = decorations.finish();
    }
    update(update) {
        if (update.docChanged) {
            this.renderWidgets();
            return;
        }
        const { state, startState } = update;
        if (update.focusChanged) {
            const line = state.doc.lineAt(state.selection.main.head);
            if (isHeading(line.text)) {
                if (update.view.hasFocus) {
                    this.decorations = this.decorations.update({
                        filterFrom: line.from,
                        filterTo: line.to,
                        filter: ()=>{
                            return false;
                        }
                    });
                } else {
                    const item = this.getDecorationForText(line);
                    if (item) {
                        this.decorations = this.decorations.update({
                            add: [
                                {
                                    from: item.from,
                                    to: item.to,
                                    value: item.value
                                }
                            ]
                        });
                    }
                }
            }
            return;
        }
        if (update.selectionSet) {
            const fromLine = startState.doc.lineAt(startState.selection.main.head);
            const toLine = state.doc.lineAt(state.selection.main.head);
            if (fromLine.number === toLine.number) {
                return;
            }
            if (isHeading(toLine.text)) {
                this.decorations = this.decorations.update({
                    filterFrom: toLine.from,
                    filterTo: toLine.to,
                    filter: ()=>{
                        return false;
                    }
                });
            }
            if (isHeading(fromLine.text)) {
                const item = this.getDecorationForText(fromLine);
                if (item) {
                    this.decorations = this.decorations.update({
                        add: [
                            {
                                from: item.from,
                                to: item.to,
                                value: item.value
                            }
                        ]
                    });
                }
            }
        }
    }
    constructor(view){
        _define_property(this, "view", void 0);
        _define_property(this, "decorations", void 0);
        this.view = view;
        this.decorations = Decoration.none;
        this.renderWidgets();
    }
}
const plugin = ViewPlugin.fromClass(SlugPlugin, {
    decorations: (v)=>v.decorations,
    provide: (v)=>{
        return EditorView.atomicRanges.of((view)=>{
            return view.plugin(v)?.decorations || Decoration.none;
        });
    }
});
export default plugin;
